import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib.gridspec import  GridSpec
from matplotlib.collections import PatchCollection
import os




# =============================================================================
# Flux calibration
# =============================================================================

def qc_stellar_extraction():
    pass
